import {
	Group,
	GROUP_DIGITS_INDEX,
	GROUP_PLACE_INDEX, GroupDigits,
	GroupPlace,
} from '../../../common';
import { numberToExponential } from '../../../exponent';
import {
	DECIMAL_POINT,
	GROUPING_SYMBOL,
	EXPONENT_DELIMITER,
	DECILLIONS_PREFIXES,
	EMPTY_GROUP_DIGITS,
	ILLION_SUFFIX,
	MILLIA_PREFIX,
	SHORT_MILLIA_DELIMITER,
	SHORT_MILLIA_ILLION_DELIMITER,
	T_AFFIX,
	THOUSAND,
	makeHundredsName,
	makeCentillionsPrefix,
	repeatString,
	StringifyGroupsOptions,
} from '../../en/common';

const getGroupName = (place: GroupPlace, shortenMillia: boolean) => {
	if (place === BigInt(0)) {
		return '' as const;
	}

	if (place === BigInt(1)) {
		return THOUSAND;
	}

	const bigGroupPlace = place - BigInt(1);
	const groupGroups = bigGroupPlace
		.toString()
		.split('')
		.reduceRight<Group[]>(
			(acc, c, i, cc) => {
				const firstGroup = acc.at(0);
				const currentPlace = BigInt(Math.floor((cc.length - i - 1) / 3));
				const newGroup = [EMPTY_GROUP_DIGITS, currentPlace] as Group;
				if (typeof firstGroup === 'undefined') {
					newGroup[GROUP_DIGITS_INDEX] = c as GroupDigits;
					return [newGroup];
				}

				if (firstGroup[0].length > 2) {
					newGroup[GROUP_DIGITS_INDEX] = c as GroupDigits;
					return [newGroup, ...acc];
				}

				newGroup[GROUP_DIGITS_INDEX] = (c + firstGroup[0]) as GroupDigits;
				return [
					newGroup,
					...acc.slice(1),
				];
			},
			[],
		)
		.map(([groupDigits, groupPlace]) => [groupDigits.padStart(3, '0'), groupPlace] as const)
		.filter(([groupDigits]) => groupDigits !== EMPTY_GROUP_DIGITS)
		.map(([groupDigits, groupPlace], _index, millias) => {
			const [hundreds, tens, ones] = groupDigits.split('').map(Number);
			const largestMillia = millias[0][GROUP_PLACE_INDEX];

			const centillionsPrefix = makeCentillionsPrefix(
				hundreds,
				tens,
				ones,
				groupPlace,
				largestMillia,
			);

			if (groupPlace < 1) {
				return centillionsPrefix;
			}

			const milliaSuffix = (
				shortenMillia && groupPlace > 1
					? `${MILLIA_PREFIX}${SHORT_MILLIA_DELIMITER}${groupPlace}${SHORT_MILLIA_ILLION_DELIMITER}`
					: repeatString(MILLIA_PREFIX, groupPlace)
			);

			if (groupDigits === '001' && groupPlace === largestMillia) {
				return milliaSuffix;
			}

			return `${centillionsPrefix}${milliaSuffix}`;
		})
		.join('');

	if (groupGroups.endsWith(DECILLIONS_PREFIXES[1])) {
		return `${groupGroups}${ILLION_SUFFIX}` as const;
	}

	if (bigGroupPlace > 10) {
		// vigin - t - illion, cen - t - illion, etc.
		return `${groupGroups}${T_AFFIX}${ILLION_SUFFIX}` as const;
	}

	return `${groupGroups}${ILLION_SUFFIX}` as const;
};

/**
 * Group a number string into groups of three digits, starting from the decimal point.
 * @param value - The number string to group.
 * @returns Group[] The groups.
 */
export const splitIntoGroups = (value: string): Group[] => {
	const [significand, exponentString] = numberToExponential(
		value,
		{
			decimalPoint: DECIMAL_POINT,
			groupingSymbol: GROUPING_SYMBOL,
			exponentDelimiter: EXPONENT_DELIMITER,
		},
	)
		.split(EXPONENT_DELIMITER);
	// FIXME use bigint for exponent and indexing???
	const exponent = Number(exponentString);
	const significantDigits = significand.replace(new RegExp(`\\${DECIMAL_POINT}`, 'g'), '');
	return significantDigits.split('').reduce<Group[]>(
		(acc, c, i) => {
			const currentPlace = BigInt(Math.floor((exponent - i) / 3));
			const lastGroup = acc.at(-1) ?? [EMPTY_GROUP_DIGITS, currentPlace];
			const currentPlaceInGroup = 2 - ((exponent - i) % 3);
			if (lastGroup[GROUP_PLACE_INDEX] === currentPlace) {
				const lastGroupDigits = lastGroup[0].split('');
				lastGroupDigits[currentPlaceInGroup] = c;
				return [...acc.slice(0, -1) ?? [], [
					lastGroupDigits.join('') as GroupDigits,
					currentPlace,
				]];
			}
			return [...acc, [c.padEnd(3, '0') as GroupDigits, currentPlace]];
		},
		[],
	);
};

/**
 * Creates a group string.
 * @param groups - The groups.
 * @param options - Options to use when creating the group.
 * @returns string[] The groups represented into strings.
 */
export const stringifyGroups = (groups: Group[], options?: StringifyGroupsOptions): string[] => {
	const filteredGroups = groups.filter(([digits, place]) => (
		place === BigInt(0) || digits !== EMPTY_GROUP_DIGITS
	));

	return filteredGroups.map(
		([group, place]) => {
			const makeHundredsArgs = group
				.padStart(3, '0')
				.split('')
				.map((s) => Number(s)) as [number, number, number];

			const groupDigitsName = makeHundredsName(
				...makeHundredsArgs,
				options?.addTensDashes ?? true,
			);
			const groupName = getGroupName(place, options?.shortenMillia ?? false);
			if (groupName.length > 0) {
				return `${groupDigitsName} ${groupName}`;
			}
			return groupDigitsName;
		},
	);
};

export {
	StringifyGroupsOptions,
	MergeTokensOptions,
	mergeTokens,
	makeNegative,
} from '../../en/common';
