import { Group, GroupPlace } from '../../common';

export const GROUP_SEPARATOR = ' ' as const;

export const DECIMAL_POINT = '.' as const;

export const GROUPING_SYMBOL = ',' as const;

export const NEGATIVE = 'negative' as const;

export const NEGATIVE_SYMBOL = '-' as const;

export const TENS_ONES_SEPARATOR = '-' as const;

export const POSITIVE_SYMBOL = '+' as const;

export const SHORT_MILLIA_DELIMITER = '^' as const;

export const SHORT_MILLIA_ILLION_DELIMITER = '-' as const;

export const EXPONENT_DELIMITER = 'e' as const;

export const EMPTY_GROUP_DIGITS = '000' as const;

export const EMPTY_PLACE: Group = [EMPTY_GROUP_DIGITS, BigInt(0)];

export const T_AFFIX = 't' as const;

/**
 * Ones number names.
 */
export const ONES = [
	'zero',
	'one',
	'two',
	'three',
	'four',
	'five',
	'six',
	'seven',
	'eight',
	'nine',
] as const;

export type OnesName = typeof ONES[number];

/**
 * Ten plus ones number names.
 */
export const TEN_PLUS_ONES = [
	'ten',
	'eleven',
	'twelve',
	'thirteen',
	'fourteen',
	'fifteen',
	'sixteen',
	'seventeen',
	'eighteen',
	'nineteen',
] as const;

export type TenPlusOnesName = typeof TEN_PLUS_ONES[number];

/**
 * Tens number names.
 */
export const TENS = [
	'zero',
	TEN_PLUS_ONES[0],
	'twenty',
	'thirty',
	'forty',
	'fifty',
	'sixty',
	'seventy',
	'eighty',
	'ninety',
] as const;

export type TensName = typeof TENS[number];

/**
 * Hundreds name.
 */
export const HUNDRED = 'hundred' as const;

/**
 * Thousands name.
 */
export const THOUSAND = 'thousand' as const;

// export const ILLION_ORDINAL_SUFFIX = 'illionth' as const;

// export const THOUSAND_ORDINAL = 'thousandth' as const;

/**
 * Special millions name.
 */
export const MILLIONS_SPECIAL_PREFIXES = [
	'',
	'm',
	'b',
	'tr',
	'quadr',
	'quint',
	'sext',
	'sept',
	'oct',
	'non',
] as const;

export type MillionsSpecialPrefix = Exclude<typeof MILLIONS_SPECIAL_PREFIXES[number], ''>;

/**
 * Millions name.
 */
export const MILLIONS_PREFIXES = [
	'',
	'un',
	'duo',
	'tre',
	'quattuor',
	'quin',
	'sex',
	'septen',
	'octo',
	'novem',
] as const;

export type MillionsPrefix = Exclude<typeof MILLIONS_PREFIXES[number], ''>;

/**
 * Decillions name.
 */
export const DECILLIONS_PREFIXES = [
	'',
	'dec',
	'vigin',
	'trigin',
	'quadragin',
	'quinquagin',
	'sexagin',
	'septuagin',
	'octogin',
	'nonagin',
] as const;

export type DecillionsPrefix = Exclude<typeof DECILLIONS_PREFIXES[number], ''>;

/**
 * Centillions name.
 */
export const CENTILLIONS_PREFIXES = [
	'',
	'cen',
	'duocen',
	'trecen',
	'quadringen',
	'quingen',
	'sescen',
	'septingen',
	'octingen',
	'nongen',
] as const;

export type CentillionsPrefix = Exclude<typeof CENTILLIONS_PREFIXES[number], ''>;

/**
 * Prefix for millia- number names.
 */
export const MILLIA_PREFIX = 'millia' as const;

/**
 * Suffix for -illion number names.
 */
export const ILLION_SUFFIX = 'illion' as const;

/**
 * Builds a name for numbers in tens and ones.
 * @param tens - Tens digit.
 * @param ones - Ones digit.
 * @param addTensDashes - Whether to add dashes between the tens and ones.
 * @returns string The name for the number.
 */
const makeTensName = (tens: number, ones: number, addTensDashes: boolean) => {
	if (tens === 0) {
		return ONES[ones];
	}

	if (tens === 1) {
		return TEN_PLUS_ONES[ones] as unknown as TenPlusOnesName;
	}

	if (ones === 0) {
		return TENS[tens];
	}

	return `${TENS[tens] as Exclude<TensName, 'zero' | 'ten'>}${addTensDashes ? TENS_ONES_SEPARATOR : ' '}${ONES[ones] as Exclude<OnesName, 'zero'>}` as const;
};

/**
 * Builds a name for numbers in hundreds, tens, and ones.
 * @param hundreds - Hundreds digit.
 * @param tens - Tens digit.
 * @param ones - Ones digit.
 * @param addTensDashes - Whether to add dashes between the tens and ones.
 * @returns string The name for the number.
 */
export const makeHundredsName = (
	hundreds: number,
	tens: number,
	ones: number,
	addTensDashes: boolean,
) => {
	if (hundreds === 0) {
		return makeTensName(tens, ones, addTensDashes);
	}

	if (tens === 0 && ones === 0) {
		return `${ONES[hundreds]} ${HUNDRED}` as const;
	}

	return `${ONES[hundreds]} ${HUNDRED} ${makeTensName(tens, ones, addTensDashes)}` as const;
};

/**
 * Builds a name for numbers in the millions.
 * @param millions - Millions digit.
 * @param currentMillia - Current millia- group.
 * @param longestMilliaCount - Number of millia- groups.
 * @returns string The millions prefix.
 */
const makeMillionsPrefix = (
	millions: number,
	currentMillia: GroupPlace,
	longestMilliaCount: GroupPlace,
) => {
	if (currentMillia === BigInt(0) && longestMilliaCount === BigInt(0)) {
		return MILLIONS_SPECIAL_PREFIXES[millions] as MillionsSpecialPrefix;
	}

	return MILLIONS_PREFIXES[millions] as MillionsPrefix;
};

/**
 * Builds a name for numbers in the decillions.
 * @param decillions - Decillions digit.
 * @param millions - Millions digit.
 * @param currentMillia - Current millia- group.
 * @param longestMilliaCount - Number of millia- groups.
 * @returns string The decillions prefix.
 */
const makeDecillionsPrefix = (
	decillions: number,
	millions: number,
	currentMillia: GroupPlace,
	longestMilliaCount: GroupPlace,
) => {
	if (decillions === 0) {
		return makeMillionsPrefix(millions, currentMillia, longestMilliaCount);
	}

	const onesPrefix = MILLIONS_PREFIXES[millions] as MillionsPrefix;
	const tensName = DECILLIONS_PREFIXES[decillions] as DecillionsPrefix;
	return `${onesPrefix}${tensName}` as const;
};

/**
 * Builds a name for numbers in the centillions.
 * @param centillions - Centillions digit.
 * @param decillions - Decillions digit.
 * @param millions - Millions digit.
 * @param currentMillia - Current millia- group.
 * @param longestMilliaCount - Number of millia- groups.
 * @returns string The centillions prefix.
 */
export const makeCentillionsPrefix = (
	centillions: number,
	decillions: number,
	millions: number,
	currentMillia: GroupPlace,
	longestMilliaCount: GroupPlace,
) => {
	if (centillions === 0) {
		return makeDecillionsPrefix(decillions, millions, currentMillia, longestMilliaCount);
	}

	const onesPrefix = MILLIONS_PREFIXES[millions] as MillionsPrefix;
	const tensName = DECILLIONS_PREFIXES[decillions] as DecillionsPrefix;
	const hundredsName = CENTILLIONS_PREFIXES[centillions] as CentillionsPrefix;
	return `${hundredsName}${onesPrefix}${decillions > 0 ? tensName : ''}` as const;
};

/**
 * Repeats a string a given number of times.
 * @param s - String to repeat.
 * @param count - Number of times to repeat the string.
 * @returns string The repeated string.
 */
export const repeatString = (s: string, count: GroupPlace) => {
	let result = '';
	for (let i = BigInt(0); i < count; i += BigInt(1)) {
		result += s;
	}
	return result;
};

export interface StringifyGroupsOptions {
	/**
	 * Whether to add dashes between tens and ones (e.g. "sixty-nine").
	 */
	addTensDashes?: boolean;
	/**
	 * Use "millia^2-tillion" instead of "milliamilliatillion".
	 */
	shortenMillia?: boolean;
}

export interface MergeTokensOptions {
	oneGroupPerLine?: boolean;
}

/**
 * Formats the final tokenized string.
 * @param tokens - The tokens to finalize.
 * @param options - The options to use.
 */
export const mergeTokens = (tokens: string[], options?: MergeTokensOptions) => (
	tokens
		.map((t) => t.trim())
		.join(options?.oneGroupPerLine ? '\n' : GROUP_SEPARATOR)
		.trim()
);

/**
 * Makes a negative string.
 * @param s - The string to make negative.
 */
export const makeNegative = (s: string) => {
	const negativePrefix = `${NEGATIVE} `;
	return s.startsWith(negativePrefix) ? s.slice(negativePrefix.length) : `${negativePrefix}${s}`;
};
